from __future__ import annotations

from typing import TYPE_CHECKING

from graphene import Boolean, Field, InputField, InputObjectType, Mutation, String
from opentelemetry import trace
from typing_extensions import Self

from infrahub.branch.merge_mutation_checker import verify_branch_merge_mutation_allowed
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.branch.enums import BranchStatus
from infrahub.database import retry_db_transaction
from infrahub.exceptions import BranchNotFoundError, ValidationError
from infrahub.graphql.context import apply_external_context
from infrahub.graphql.field_extractor import extract_graphql_fields
from infrahub.graphql.types.context import ContextInput
from infrahub.log import get_logger
from infrahub.workflows.catalogue import (
    BRANCH_CREATE,
    BRANCH_DELETE,
    BRANCH_MERGE_MUTATION,
    BRANCH_REBASE,
    BRANCH_VALIDATE,
)

from ..types import BranchType
from ..types.task import TaskInfo
from .models import BranchCreateModel

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from ..initialization import GraphqlContext


log = get_logger()


class BranchCreateInput(InputObjectType):
    id = String(required=False)
    name = String(required=True)
    description = String(required=False)
    origin_branch = String(required=False)
    branched_from = String(required=False)
    sync_with_git = Boolean(required=False)
    is_isolated = InputField(Boolean(required=False), deprecation_reason="Non isolated mode is not supported anymore")


class BranchCreate(Mutation):
    class Arguments:
        data = BranchCreateInput(required=True)
        context = ContextInput(required=False)
        background_execution = Boolean(required=False, deprecation_reason="Please use `wait_until_completion` instead")
        wait_until_completion = Boolean(required=False)

    ok = Boolean()
    object = Field(BranchType)
    task = Field(TaskInfo, required=False)

    @classmethod
    @trace.get_tracer(__name__).start_as_current_span("branch_create")
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: BranchCreateInput,
        context: ContextInput | None = None,
        background_execution: bool = False,
        wait_until_completion: bool = True,
    ) -> Self:
        if data.origin_branch and data.origin_branch != registry.default_branch:
            raise ValueError(f"origin_branch must be '{registry.default_branch}'")

        graphql_context: GraphqlContext = info.context
        task: dict | None = None

        model = BranchCreateModel(**data)
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        try:
            await Branch.get_by_name(db=graphql_context.db, name=model.name)
            raise ValidationError(f"The branch {model.name} already exists")
        except BranchNotFoundError:
            pass

        if background_execution or not wait_until_completion:
            workflow = await graphql_context.active_service.workflow.submit_workflow(
                workflow=BRANCH_CREATE, context=graphql_context.get_context(), parameters={"model": model}
            )
            task = {"id": workflow.id}
            return cls(ok=True, task=task)

        await graphql_context.active_service.workflow.execute_workflow(
            workflow=BRANCH_CREATE, context=graphql_context.get_context(), parameters={"model": model}
        )

        # Retrieve created branch
        obj = await Branch.get_by_name(db=graphql_context.db, name=model.name)
        fields = extract_graphql_fields(info=info)
        return cls(object=await obj.to_graphql(fields=fields.get("object", {})), ok=True, task=task)


class BranchNameInput(InputObjectType):
    name = String(required=False)


class BranchUpdateInput(InputObjectType):
    name = String(required=True)
    description = String(required=False)
    is_isolated = InputField(Boolean(required=False), deprecation_reason="Non isolated mode is not supported anymore")


class BranchDelete(Mutation):
    class Arguments:
        data = BranchNameInput(required=True)
        context = ContextInput(required=False)
        wait_until_completion = Boolean(required=False)

    ok = Boolean()
    task = Field(TaskInfo, required=False)

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: BranchNameInput,
        context: ContextInput | None = None,
        wait_until_completion: bool = True,
    ) -> Self:
        graphql_context: GraphqlContext = info.context
        obj = await Branch.get_by_name(db=graphql_context.db, name=str(data.name))
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        if wait_until_completion:
            await graphql_context.active_service.workflow.execute_workflow(
                workflow=BRANCH_DELETE, context=graphql_context.get_context(), parameters={"branch": obj.name}
            )
            return cls(ok=True)

        workflow = await graphql_context.active_service.workflow.submit_workflow(
            workflow=BRANCH_DELETE, context=graphql_context.get_context(), parameters={"branch": obj.name}
        )
        return cls(ok=True, task={"id": str(workflow.id)})


class BranchUpdate(Mutation):
    class Arguments:
        data = BranchUpdateInput(required=True)
        context = ContextInput(required=False)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="branch_update")
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: BranchNameInput,
        context: ContextInput | None = None,
    ) -> Self:
        graphql_context: GraphqlContext = info.context

        obj = await Branch.get_by_name(db=graphql_context.db, name=data["name"])
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        to_extract = ["description"]
        for field_name in to_extract:
            if field_name in data and data.get(field_name) is not None:
                setattr(obj, field_name, data[field_name])

        async with graphql_context.db.start_transaction() as db:
            await obj.save(db=db)

        return cls(ok=True)


class BranchRebase(Mutation):
    class Arguments:
        data = BranchNameInput(required=True)
        context = ContextInput(required=False)
        wait_until_completion = Boolean(required=False)

    ok = Boolean()
    object = Field(BranchType)
    task = Field(TaskInfo, required=False)

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: BranchNameInput,
        context: ContextInput | None = None,
        wait_until_completion: bool = True,
    ) -> Self:
        graphql_context: GraphqlContext = info.context

        obj = await Branch.get_by_name(db=graphql_context.db, name=str(data.name))
        await apply_external_context(graphql_context=graphql_context, context_input=context)
        task: dict | None = None

        if wait_until_completion:
            await graphql_context.active_service.workflow.execute_workflow(
                workflow=BRANCH_REBASE, context=graphql_context.get_context(), parameters={"branch": obj.name}
            )

            # Pull the latest information about the branch from the database directly
            obj = await Branch.get_by_name(db=graphql_context.db, name=str(data.name))
        else:
            workflow = await graphql_context.active_service.workflow.submit_workflow(
                workflow=BRANCH_REBASE, context=graphql_context.get_context(), parameters={"branch": obj.name}
            )
            task = {"id": workflow.id}

        fields = extract_graphql_fields(info=info)
        ok = True

        return cls(object=await obj.to_graphql(fields=fields.get("object", {})), ok=ok, task=task)


class BranchValidate(Mutation):
    class Arguments:
        data = BranchNameInput(required=True)
        context = ContextInput(required=False)
        wait_until_completion = Boolean(required=False)

    ok = Boolean()
    object = Field(BranchType)
    task = Field(TaskInfo, required=False)

    @classmethod
    @retry_db_transaction(name="branch_validate")
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: BranchNameInput,
        context: ContextInput | None = None,
        wait_until_completion: bool = True,
    ) -> Self:
        graphql_context: GraphqlContext = info.context

        obj = await Branch.get_by_name(db=graphql_context.db, name=str(data.name))
        await apply_external_context(graphql_context=graphql_context, context_input=context)
        task: dict | None = None
        ok = True

        if wait_until_completion:
            await graphql_context.active_service.workflow.execute_workflow(
                workflow=BRANCH_VALIDATE, context=graphql_context.get_context(), parameters={"branch": obj.name}
            )
        else:
            workflow = await graphql_context.active_service.workflow.submit_workflow(
                workflow=BRANCH_VALIDATE, context=graphql_context.get_context(), parameters={"branch": obj.name}
            )
            task = {"id": workflow.id}

        fields = extract_graphql_fields(info=info)

        return cls(object=await obj.to_graphql(fields=fields.get("object", {})), ok=ok, task=task)


class BranchMerge(Mutation):
    class Arguments:
        data = BranchNameInput(required=True)
        context = ContextInput(required=False)
        wait_until_completion = Boolean(required=False)

    ok = Boolean()
    object = Field(BranchType)
    task = Field(TaskInfo, required=False)

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: BranchNameInput,
        context: ContextInput | None = None,
        wait_until_completion: bool = True,
    ) -> Self:
        branch_name = data["name"]
        task: dict | None = None
        graphql_context: GraphqlContext = info.context
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        await verify_branch_merge_mutation_allowed(
            db=graphql_context.db, account_session=graphql_context.active_account_session
        )

        obj = await Branch.get_by_name(db=graphql_context.db, name=branch_name)
        if obj.status == BranchStatus.NEED_UPGRADE_REBASE:
            raise ValidationError(f"Cannot merge branch '{branch_name}' with status '{obj.status.name}'")

        if wait_until_completion:
            await graphql_context.active_service.workflow.execute_workflow(
                workflow=BRANCH_MERGE_MUTATION,
                context=graphql_context.get_context(),
                parameters={"branch": branch_name},
            )
        else:
            workflow = await graphql_context.active_service.workflow.submit_workflow(
                workflow=BRANCH_MERGE_MUTATION,
                context=graphql_context.get_context(),
                parameters={"branch": branch_name},
            )
            task = {"id": workflow.id}

        # Pull the latest information about the branch from the database directly
        obj = await Branch.get_by_name(db=graphql_context.db, name=branch_name)

        fields = extract_graphql_fields(info=info)
        ok = True

        return cls(object=await obj.to_graphql(fields=fields.get("object", {})), ok=ok, task=task)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from infrahub.core.timestamp import Timestamp
from infrahub.exceptions import InitializationError

if TYPE_CHECKING:
    import graphene

    from infrahub.core.branch import Branch
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.graphql.manager import GraphQLSchemaManager, InterfaceReference

    from .mutations.main import InfrahubMutation
    from .types import InfrahubObject


@dataclass
class BranchDetails:
    schema_changed_at: Timestamp
    schema_hash: str
    gql_manager: GraphQLSchemaManager


@dataclass
class GraphQLSchemaRegistry:
    _branch_details_by_hash: dict[str, BranchDetails] = field(default_factory=dict)
    _branch_name_by_hash: dict[str, set[str]] = field(default_factory=dict)
    _branch_hash_activation_by_branch_name: dict[str, dict[str, str]] = field(default_factory=dict)
    _registered_interface_types: dict[str, type[graphene.Interface]] = field(default_factory=dict)
    _reference_hash_schema_map: dict[str, set[str]] = field(default_factory=dict)
    _registered_edge_types: dict[str, type[InfrahubObject]] = field(default_factory=dict)
    _registered_paginated_types: dict[str, type[InfrahubObject]] = field(default_factory=dict)
    _registered_input_types: dict[str, type[graphene.InputObjectType]] = field(default_factory=dict)
    _registered_object_types: dict[str, type[InfrahubObject]] = field(default_factory=dict)
    _registered_mutation_types: dict[str, type[InfrahubMutation]] = field(default_factory=dict)

    _manager_class: type[GraphQLSchemaManager] | None = None

    def _add_branch_hash(self, branch_name: str, schema_hash: str) -> None:
        if schema_hash not in self._branch_name_by_hash:
            self._branch_name_by_hash[schema_hash] = set()

        self._branch_name_by_hash[schema_hash].add(branch_name)

    def _register_manager(self, manager: type[GraphQLSchemaManager]) -> None:
        self._manager_class = manager

    @property
    def manager(self) -> type[GraphQLSchemaManager]:
        if self._manager_class:
            return self._manager_class
        raise InitializationError

    def clear_cache(self) -> None:
        """Clear internal cache stored within this registry."""
        self._branch_details_by_hash = {}
        self._branch_name_by_hash = {}
        self._branch_hash_activation_by_branch_name = {}
        self._registered_interface_types = {}
        self._reference_hash_schema_map = {}
        self._registered_edge_types = {}
        self._registered_paginated_types = {}
        self._registered_input_types = {}
        self._registered_object_types = {}
        self._registered_mutation_types = {}

    def _add_schema_to_reference_hash(self, reference_hash: str, schema_hash: str) -> None:
        """Add the schema hash to a map containing the referenced object.

        The goal of this is to be able to see all schemas that use a given reference type,
        once no schemas use a specific type it's safe to remove the type from the registry.
        """
        if reference_hash not in self._reference_hash_schema_map:
            self._reference_hash_schema_map[reference_hash] = set()
        self._reference_hash_schema_map[reference_hash].add(schema_hash)

    def get_edge_type(self, reference_hash: str, schema_hash: str) -> type[InfrahubObject] | None:
        self._add_schema_to_reference_hash(reference_hash=reference_hash, schema_hash=schema_hash)
        return self._registered_edge_types.get(reference_hash)

    def set_edge_type(self, reference: type[InfrahubObject], reference_hash: str, schema_hash: str) -> None:
        self._add_schema_to_reference_hash(reference_hash=reference_hash, schema_hash=schema_hash)
        self._registered_edge_types[reference_hash] = reference

    def get_input_type(self, reference_hash: str, schema_hash: str) -> type[graphene.InputObjectType] | None:
        self._add_schema_to_reference_hash(reference_hash=reference_hash, schema_hash=schema_hash)
        return self._registered_input_types.get(reference_hash)

    def set_input_type(self, reference: type[graphene.InputObjectType], reference_hash: str, schema_hash: str) -> None:
        self._add_schema_to_reference_hash(reference_hash=reference_hash, schema_hash=schema_hash)
        self._registered_input_types[reference_hash] = reference

    def get_interface_type(self, reference_hash: str, schema_hash: str) -> type[graphene.Interface] | None:
        self._add_schema_to_reference_hash(reference_hash=reference_hash, schema_hash=schema_hash)
        return self._registered_interface_types.get(reference_hash)

    def set_interface_type(self, reference: InterfaceReference, schema_hash: str) -> None:
        self._add_schema_to_reference_hash(reference_hash=reference.reference_hash, schema_hash=schema_hash)
        self._registered_interface_types[reference.reference_hash] = reference.reference

    def get_mutation_type(self, reference_hash: str, schema_hash: str) -> type[InfrahubMutation] | None:
        self._add_schema_to_reference_hash(reference_hash=reference_hash, schema_hash=schema_hash)
        return self._registered_mutation_types.get(reference_hash)

    def set_mutation_type(self, reference: type[InfrahubMutation], reference_hash: str, schema_hash: str) -> None:
        self._add_schema_to_reference_hash(reference_hash=reference_hash, schema_hash=schema_hash)
        self._registered_mutation_types[reference_hash] = reference

    def get_object_type(self, reference_hash: str, schema_hash: str) -> type[InfrahubObject] | None:
        self._add_schema_to_reference_hash(reference_hash=reference_hash, schema_hash=schema_hash)
        return self._registered_object_types.get(reference_hash)

    def set_object_type(self, reference: type[InfrahubObject], reference_hash: str, schema_hash: str) -> None:
        self._add_schema_to_reference_hash(reference_hash=reference_hash, schema_hash=schema_hash)
        self._registered_object_types[reference_hash] = reference

    def get_paginated_type(self, reference_hash: str, schema_hash: str) -> type[InfrahubObject] | None:
        self._add_schema_to_reference_hash(reference_hash=reference_hash, schema_hash=schema_hash)
        return self._registered_paginated_types.get(reference_hash)

    def set_paginated_type(self, reference: type[InfrahubObject], reference_hash: str, schema_hash: str) -> None:
        self._add_schema_to_reference_hash(reference_hash=reference_hash, schema_hash=schema_hash)
        self._registered_paginated_types[reference_hash] = reference

    def purge_inactive(self, active_branches: list[str]) -> set[str]:
        """Return inactive branches that were purged"""
        inactive_branches: set[str] = set()
        for schema_hash in list(self._branch_name_by_hash.keys()):
            branches = list(self._branch_name_by_hash[schema_hash])
            for branch in branches:
                if branch not in active_branches and branch in self._branch_name_by_hash[schema_hash]:
                    inactive_branches.add(branch)
                    self._branch_name_by_hash[schema_hash].discard(branch)

        for schema_hash in list(self._branch_name_by_hash.keys()):
            if not self._branch_name_by_hash[schema_hash]:
                # If no remaining branch is using the schema remove it completely
                del self._branch_name_by_hash[schema_hash]
                del self._branch_details_by_hash[schema_hash]

        return inactive_branches

    def cache_branch(self, branch: Branch, schema_branch: SchemaBranch, schema_hash: str) -> BranchDetails:
        branch_details = BranchDetails(
            schema_changed_at=Timestamp(branch.schema_changed_at) if branch.schema_changed_at else Timestamp(),
            schema_hash=schema_hash,
            gql_manager=self.manager(schema=schema_branch),
        )

        self._branch_details_by_hash[schema_hash] = branch_details

        return branch_details

    def get_manager_for_branch(self, branch: Branch, schema_branch: SchemaBranch) -> GraphQLSchemaManager:
        if branch.schema_hash:
            schema_hash = branch.schema_hash.main
        else:
            schema_hash = schema_branch.get_hash()

        if schema_hash in self._branch_details_by_hash:
            branch_details = self._branch_details_by_hash[schema_hash]
        else:
            branch_details = self.cache_branch(branch=branch, schema_branch=schema_branch, schema_hash=schema_hash)

        self._add_branch_hash(branch_name=branch.name, schema_hash=schema_hash)

        return branch_details.gql_manager


registry = GraphQLSchemaRegistry()



from dataclasses import asdict

from infrahub.core.query import QueryType
from infrahub.database import InfrahubDatabase

from .models import EdgeToUpdate


class PatchPlanEdgeUpdater:
    def __init__(self, db: InfrahubDatabase, batch_size_limit: int = 1000) -> None:
        self.db = db
        self.batch_size_limit = batch_size_limit

    async def _run_update_query(self, edges_to_update: list[EdgeToUpdate]) -> None:
        query = """
UNWIND $edges_to_update AS edge_to_update
MATCH ()-[e]-()
WHERE %(id_func_name)s(e) = edge_to_update.db_id
SET e = edge_to_update.after_props
        """ % {"id_func_name": self.db.get_id_function_name()}
        await self.db.execute_query(
            query=query, params={"edges_to_update": [asdict(e) for e in edges_to_update]}, type=QueryType.WRITE
        )

    async def execute(self, edges_to_update: list[EdgeToUpdate]) -> None:
        for i in range(0, len(edges_to_update), self.batch_size_limit):
            vertices_slice = edges_to_update[i : i + self.batch_size_limit]
            await self._run_update_query(edges_to_update=vertices_slice)



from infrahub.core.registry import registry
from infrahub.exceptions import SchemaNotFoundError


def get_branches_with_schema_number_pool(kind: str, attribute_name: str) -> list[str]:
    """Return branches where schema defined NumberPool exists"""

    registered_branches = []
    active_branches = registry.schema.get_branches()

    for active_branch in active_branches:
        try:
            schema = registry.schema.get(name=kind, branch=active_branch)
        except SchemaNotFoundError:
            continue

        if attribute_name in schema.attribute_names:
            attribute = schema.get_attribute(name=attribute_name)
            if attribute.kind == "NumberPool":
                registered_branches.append(active_branch)

    return registered_branches



from dataclasses import dataclass

from infrahub.core.branch import Branch
from infrahub.core.changelog.models import RelationshipCardinalityManyChangelog
from infrahub.core.constants import DiffAction, InfrahubKind
from infrahub.core.schema import NodeSchema
from infrahub.database import InfrahubDatabase
from infrahub.events.group_action import GroupMemberAddedEvent, GroupMemberRemovedEvent, GroupMutatedEvent
from infrahub.events.models import EventMeta, EventNode, InfrahubEvent
from infrahub.events.node_action import NodeCreatedEvent, NodeMutatedEvent, NodeUpdatedEvent
from infrahub.exceptions import SchemaNotFoundError

from .ancestors import collect_ancestors


@dataclass
class ApplicableEvent:
    event: NodeMutatedEvent
    node_schema: NodeSchema
    relationship: RelationshipCardinalityManyChangelog


class GroupNodeMutationParser:
    def __init__(self, db: InfrahubDatabase, branch: Branch) -> None:
        self._db = db
        self._branch = branch

    def _get_schema(self, kind: str) -> NodeSchema | None:
        try:
            node_schema = self._db.schema.get_node_schema(name=kind, branch=self._branch, duplicate=False)
        except (ValueError, SchemaNotFoundError):
            return None
        return node_schema

    def _get_applicable_events(self, events: list[NodeMutatedEvent]) -> list[ApplicableEvent]:
        applicable: list[ApplicableEvent] = []
        for event in events:
            if event_kind := self._get_schema(kind=event.kind):
                if (
                    InfrahubKind.GENERICGROUP in event_kind.inherit_from
                    and "members" in event.changelog.relationships
                    and isinstance(
                        event.changelog.relationships["members"],
                        RelationshipCardinalityManyChangelog,
                    )
                ):
                    applicable.append(
                        ApplicableEvent(
                            event=event,
                            node_schema=event_kind,
                            relationship=event.changelog.relationships["members"],
                        )
                    )
        return applicable

    async def group_events_from_node_actions(self, events: list[NodeMutatedEvent]) -> list[InfrahubEvent]:
        group_events: list[InfrahubEvent] = []

        for applicable_event in self._get_applicable_events(events=events):
            added_peers = [
                EventNode(id=peer.peer_id, kind=peer.peer_kind)
                for peer in applicable_event.relationship.peers
                if peer.peer_status == DiffAction.ADDED
            ]
            removed_peers = [
                EventNode(id=peer.peer_id, kind=peer.peer_kind)
                for peer in applicable_event.relationship.peers
                if peer.peer_status == DiffAction.REMOVED
            ]
            if added_peers:
                group_events.append(
                    await self._define_group_event(
                        event_type=GroupMemberAddedEvent, parent=applicable_event.event, peers=added_peers
                    )
                )

            if removed_peers:
                group_events.append(
                    await self._define_group_event(
                        event_type=GroupMemberRemovedEvent, parent=applicable_event.event, peers=removed_peers
                    )
                )

        return group_events

    async def _define_group_event(
        self,
        event_type: type[GroupMemberAddedEvent | GroupMemberRemovedEvent],
        parent: NodeMutatedEvent,
        peers: list[EventNode],
    ) -> GroupMutatedEvent:
        event_meta = EventMeta.from_parent(parent=parent)
        group_event = event_type(
            meta=event_meta,
            kind=parent.kind,
            node_id=parent.node_id,
            members=peers,
        )
        if isinstance(parent, NodeCreatedEvent | NodeUpdatedEvent):
            # Avoid trying to find ancestors for deleted nodes
            group_event.ancestors = await collect_ancestors(
                db=self._db,
                branch=self._branch,
                node_kind=parent.kind,
                node_id=parent.node_id,
            )
        return group_event



from infrahub.core.validators.node.inherit_from import NodeInheritFromChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class SchemaInheritFromConstraintDependency(DependencyBuilder[NodeInheritFromChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> NodeInheritFromChecker:
        return NodeInheritFromChecker(db=context.db, branch=context.branch)



from typing import ClassVar

from pydantic import Field

from infrahub.core.constants import InfrahubKind, MutationAction

from .constants import EVENT_NAMESPACE
from .models import EventNode, InfrahubEvent


class GroupMutatedEvent(InfrahubEvent):
    """Event generated when a node has been mutated"""

    kind: str = Field(..., description="The type of updated group")
    node_id: str = Field(..., description="The ID of the updated group")
    action: MutationAction = Field(..., description="The action taken on the node")
    members: list[EventNode] = Field(default_factory=list, description="Updated members during this event.")
    ancestors: list[EventNode] = Field(
        default_factory=list, description="A list of groups that are ancestors of this group."
    )

    def get_related(self) -> list[dict[str, str]]:
        related = super().get_related()

        if self.kind in [InfrahubKind.GENERATORGROUP, InfrahubKind.GENERATORAWAREGROUP, InfrahubKind.GRAPHQLQUERYGROUP]:
            # Temporary workaround to avoid too large payloads for the related field
            return related

        related.append(
            {
                "prefect.resource.id": self.node_id,
                "prefect.resource.role": "infrahub.related.node",
                "infrahub.node.kind": self.kind,
            }
        )
        related.append(
            {
                "prefect.resource.id": self.node_id,
                "prefect.resource.role": "infrahub.group.update",
                "infrahub.node.kind": self.kind,
            }
        )

        for member in self.members:
            related.append(
                {
                    "prefect.resource.id": member.id,
                    "prefect.resource.role": "infrahub.group.member",
                    "infrahub.node.kind": member.kind,
                }
            )
            related.append(
                {
                    "prefect.resource.id": member.id,
                    "prefect.resource.role": "infrahub.related.node",
                    "infrahub.node.kind": member.kind,
                }
            )

        for ancestor in self.ancestors:
            related.append(
                {
                    "prefect.resource.id": ancestor.id,
                    "prefect.resource.role": "infrahub.group.ancestor",
                    "infrahub.node.kind": ancestor.kind,
                }
            )
            related.append(
                {
                    "prefect.resource.id": ancestor.id,
                    "prefect.resource.role": "infrahub.related.node",
                    "infrahub.node.kind": ancestor.kind,
                }
            )
            related.append(
                {
                    "prefect.resource.id": ancestor.id,
                    "prefect.resource.role": "infrahub.group.update",
                    "infrahub.node.kind": ancestor.kind,
                }
            )

        return related

    def get_resource(self) -> dict[str, str]:
        return {
            "prefect.resource.id": f"infrahub.node.{self.node_id}",
            "infrahub.node.kind": self.kind,
            "infrahub.node.id": self.node_id,
            "infrahub.node.action": self.action.value,
            "infrahub.node.root_id": self.node_id,
            "infrahub.branch.name": self.meta.context.branch.name,
        }


class GroupMemberAddedEvent(GroupMutatedEvent):
    """Event generated when a one or more members have been added to a group"""

    action: MutationAction = MutationAction.CREATED
    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.group.member_added"
    infrahub_node_kind_event: ClassVar[bool] = True


class GroupMemberRemovedEvent(GroupMutatedEvent):
    """Event generated when a one or more members have been removed to a group"""

    action: MutationAction = MutationAction.DELETED
    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.group.member_removed"
    infrahub_node_kind_event: ClassVar[bool] = True



from __future__ import annotations

from prefect import flow
from prefect.logging import get_run_logger

from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.core.registry import registry
from infrahub.pools.tasks import validate_schema_number_pools
from infrahub.services import InfrahubServices  # noqa: TC001  needed for prefect flow
from infrahub.workflows.utils import wait_for_schema_to_converge


@flow(
    name="branch-merged",
    flow_run_name="Running actions after '{source_branch}' was merged",
)
async def branch_merged(
    source_branch: str,  # noqa: ARG001
    context: InfrahubContext,
    service: InfrahubServices,
    target_branch: str | None = None,
) -> None:
    target_branch = target_branch or registry.default_branch
    log = get_run_logger()
    await wait_for_schema_to_converge(
        branch_name=target_branch, component=service.component, db=service.database, log=log
    )

    await validate_schema_number_pools(branch_name=target_branch, context=context, service=service)



from __future__ import annotations

import ast
from collections import defaultdict
from pathlib import Path
from typing import Optional

import typer
from ariadne_codegen.client_generators.package import PackageGenerator, get_package_generator
from ariadne_codegen.exceptions import ParsingError
from ariadne_codegen.plugins.explorer import get_plugins_types
from ariadne_codegen.plugins.manager import PluginManager
from ariadne_codegen.schema import (
    filter_fragments_definitions,
    filter_operations_definitions,
    get_graphql_schema_from_path,
)
from ariadne_codegen.settings import ClientSettings, CommentsStrategy
from ariadne_codegen.utils import ast_to_str
from graphql import DefinitionNode, GraphQLSchema, NoUnusedFragmentsRule, parse, specified_rules, validate
from rich.console import Console

from ..async_typer import AsyncTyper
from ..ctl.client import initialize_client
from ..ctl.utils import catch_exception
from ..graphql.utils import insert_fragments_inline, remove_fragment_import
from .parameters import CONFIG_PARAM

app = AsyncTyper()
console = Console()

ARIADNE_PLUGINS = [
    "infrahub_sdk.graphql.plugin.PydanticBaseModelPlugin",
    "infrahub_sdk.graphql.plugin.FutureAnnotationPlugin",
    "infrahub_sdk.graphql.plugin.StandardTypeHintPlugin",
]


def find_gql_files(query_path: Path) -> list[Path]:
    """
    Find all files with .gql extension in the specified directory.

    Args:
        query_path: Path to the directory to search for .gql files

    Returns:
        List of Path objects for all .gql files found
    """
    if not query_path.exists():
        raise FileNotFoundError(f"File or directory not found: {query_path}")

    if not query_path.is_dir() and query_path.is_file():
        return [query_path]

    return list(query_path.glob("**/*.gql"))


def get_graphql_query(queries_path: Path, schema: GraphQLSchema) -> tuple[DefinitionNode, ...]:
    """Get GraphQL queries definitions from a single GraphQL file."""

    if not queries_path.exists():
        raise FileNotFoundError(f"File not found: {queries_path}")
    if not queries_path.is_file():
        raise ValueError(f"{queries_path} is not a file")

    queries_str = queries_path.read_text(encoding="utf-8")
    queries_ast = parse(queries_str)
    validation_errors = validate(
        schema=schema,
        document_ast=queries_ast,
        rules=[r for r in specified_rules if r is not NoUnusedFragmentsRule],
    )
    if validation_errors:
        raise ValueError("\n\n".join(error.message for error in validation_errors))
    return queries_ast.definitions


def generate_result_types(directory: Path, package: PackageGenerator, fragment: ast.Module) -> None:
    for file_name, module in package._result_types_files.items():
        file_path = directory / file_name

        insert_fragments_inline(module, fragment)
        remove_fragment_import(module)

        code = package._add_comments_to_code(ast_to_str(module), package.queries_source)
        if package.plugin_manager:
            code = package.plugin_manager.generate_result_types_code(code)
        file_path.write_text(code)
        package._generated_files.append(file_path.name)


@app.callback()
def callback() -> None:
    """
    Various GraphQL related commands.
    """


@app.command()
@catch_exception(console=console)
async def export_schema(
    destination: Path = typer.Option("schema.graphql", help="Path to the GraphQL schema file."),
    _: str = CONFIG_PARAM,
) -> None:
    """Export the GraphQL schema to a file."""

    client = initialize_client()
    schema_text = await client.schema.get_graphql_schema()

    destination.parent.mkdir(parents=True, exist_ok=True)
    destination.write_text(schema_text)
    console.print(f"[green]Schema exported to {destination}")


@app.command()
@catch_exception(console=console)
async def generate_return_types(
    query: Optional[Path] = typer.Argument(
        None, help="Location of the GraphQL query file(s). Defaults to current directory if not specified."
    ),
    schema: Path = typer.Option("schema.graphql", help="Path to the GraphQL schema file."),
    _: str = CONFIG_PARAM,
) -> None:
    """Create Pydantic Models for GraphQL query return types"""

    query = Path.cwd() if query is None else query

    # Load the GraphQL schema
    if not schema.exists():
        raise FileNotFoundError(f"GraphQL Schema file not found: {schema}")
    graphql_schema = get_graphql_schema_from_path(schema_path=str(schema))

    # Initialize the plugin manager
    plugin_manager = PluginManager(
        schema=graphql_schema,
        plugins_types=get_plugins_types(plugins_strs=ARIADNE_PLUGINS),
    )

    # Find the GraphQL files and organize them by directory
    gql_files = find_gql_files(query)
    gql_per_directory: dict[Path, list[Path]] = defaultdict(list)
    for gql_file in gql_files:
        gql_per_directory[gql_file.parent].append(gql_file)

    # Generate the Pydantic Models for the GraphQL queries
    for directory, gql_files in gql_per_directory.items():
        for gql_file in gql_files:
            try:
                definitions = get_graphql_query(queries_path=gql_file, schema=graphql_schema)
            except ValueError as exc:
                console.print(f"[red]Error generating result types for {gql_file}: {exc}")
                continue
            queries = filter_operations_definitions(definitions)
            fragments = filter_fragments_definitions(definitions)

            package_generator = get_package_generator(
                schema=graphql_schema,
                fragments=fragments,
                settings=ClientSettings(
                    schema_path=str(schema),
                    target_package_name=directory.name,
                    queries_path=str(directory),
                    include_comments=CommentsStrategy.NONE,
                ),
                plugin_manager=plugin_manager,
            )

            parsing_failed = False
            try:
                for query_operation in queries:
                    package_generator.add_operation(query_operation)
            except ParsingError as exc:
                console.print(f"[red]Unable to process {gql_file.name}: {exc}")
                parsing_failed = True

            if parsing_failed:
                continue

            module_fragment = package_generator.fragments_generator.generate()

            generate_result_types(directory=directory, package=package_generator, fragment=module_fragment)

            for file_name in package_generator._result_types_files.keys():
                console.print(f"[green]Generated {file_name} in {directory}")



from infrahub.core.diff.combiner import DiffCombiner
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class DiffCombinerDependency(DependencyBuilder[DiffCombiner]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> DiffCombiner:  # noqa: ARG003
        return DiffCombiner()



from __future__ import annotations

from graphene import Field, List, NonNull, ObjectType, String


class RelationshipPeer(ObjectType):
    id = String(required=False)
    kind = String(required=False)


class Relationship(ObjectType):
    id = String(required=False)
    identifier = String(required=False)
    peers = List(NonNull(RelationshipPeer))


class RelationshipNode(ObjectType):
    node = Field(Relationship, required=True)



from typing import TYPE_CHECKING, Awaitable, Callable

from prefect import get_run_logger, task
from prefect.automations import AutomationCore
from prefect.cache_policies import NONE
from prefect.client.orchestration import PrefectClient, get_client
from prefect.client.schemas.filters import DeploymentFilter, DeploymentFilterName
from prefect.events.schemas.automations import Automation
from prefect.exceptions import PrefectHTTPStatusError

from infrahub import lock
from infrahub.database import InfrahubDatabase
from infrahub.trigger.models import TriggerDefinition

from .models import TriggerSetupReport, TriggerType

if TYPE_CHECKING:
    from uuid import UUID


def compare_automations(target: AutomationCore, existing: Automation) -> bool:
    """Compare an AutomationCore with an existing Automation object to identify if they are identical or not

    Return True if the target is identical to the existing automation
    """

    target_dump = target.model_dump(exclude_defaults=True, exclude_none=True)
    existing_dump = existing.model_dump(exclude_defaults=True, exclude_none=True, exclude={"id"})

    return target_dump == existing_dump


@task(name="trigger-setup-specific", task_run_name="Setup triggers of a specific kind", cache_policy=NONE)  # type: ignore[arg-type]
async def setup_triggers_specific(
    gatherer: Callable[[InfrahubDatabase | None], Awaitable[list[TriggerDefinition]]],
    trigger_type: TriggerType,
    db: InfrahubDatabase | None = None,
) -> TriggerSetupReport:
    async with lock.registry.get(
        name=f"configure-action-rules-{trigger_type.value}", namespace="trigger-rules", local=False
    ):
        if db:
            async with db.start_session(read_only=True) as dbs:
                triggers = await gatherer(dbs)
        else:
            triggers = await gatherer(db)
        async with get_client(sync_client=False) as prefect_client:
            return await setup_triggers(
                client=prefect_client,
                triggers=triggers,
                trigger_type=trigger_type,
            )  # type: ignore[misc]


@task(name="trigger-setup", task_run_name="Setup triggers", cache_policy=NONE)
async def setup_triggers(
    client: PrefectClient,
    triggers: list[TriggerDefinition],
    trigger_type: TriggerType | None = None,
    force_update: bool = False,
) -> TriggerSetupReport:
    log = get_run_logger()

    report = TriggerSetupReport()

    if trigger_type:
        log.debug(f"Setting up triggers of type {trigger_type.value}")
    else:
        log.debug("Setting up all triggers")

    # -------------------------------------------------------------
    # Retrieve existing Deployments and Automation from the server
    # -------------------------------------------------------------
    deployment_names = list({name for trigger in triggers for name in trigger.get_deployment_names()})
    deployments = {
        item.name: item
        for item in await client.read_deployments(
            deployment_filter=DeploymentFilter(name=DeploymentFilterName(any_=deployment_names))
        )
    }
    deployments_mapping: dict[str, UUID] = {name: item.id for name, item in deployments.items()}

    # If a trigger type is provided, narrow down the list of existing triggers to know which one to delete
    existing_automations: dict[str, Automation] = {}
    if trigger_type:
        existing_automations = {
            item.name: item
            for item in await client.read_automations()
            if item.name.startswith(f"{trigger_type.value}::")
        }
    else:
        existing_automations = {item.name: item for item in await client.read_automations()}

    trigger_names = [trigger.generate_name() for trigger in triggers]
    automation_names = list(existing_automations.keys())

    log.debug(f"{len(automation_names)} existing triggers ({automation_names})")
    log.debug(f"{len(trigger_names)} triggers to configure ({trigger_names})")

    to_delete = set(automation_names) - set(trigger_names)
    log.debug(f"{len(to_delete)} triggers to delete ({to_delete})")

    # -------------------------------------------------------------
    # Create or Update all triggers
    # -------------------------------------------------------------
    for trigger in triggers:
        automation = AutomationCore(
            name=trigger.generate_name(),
            description=trigger.get_description(),
            enabled=True,
            trigger=trigger.trigger.get_prefect(),
            actions=[action.get_prefect(mapping=deployments_mapping) for action in trigger.actions],
        )

        existing_automation = existing_automations.get(trigger.generate_name(), None)

        if existing_automation:
            if force_update or not compare_automations(target=automation, existing=existing_automation):
                await client.update_automation(automation_id=existing_automation.id, automation=automation)
                log.info(f"{trigger.generate_name()} Updated")
                report.updated.append(trigger)
            else:
                report.unchanged.append(trigger)
        else:
            await client.create_automation(automation=automation)
            log.info(f"{trigger.generate_name()} Created")
            report.created.append(trigger)

    # -------------------------------------------------------------
    # Delete Triggers that shouldn't be there
    # -------------------------------------------------------------
    for item_to_delete in to_delete:
        existing_automation = existing_automations.get(item_to_delete)

        if not existing_automation:
            continue

        report.deleted.append(existing_automation)
        try:
            await client.delete_automation(automation_id=existing_automation.id)
            log.info(f"{item_to_delete} Deleted")
        except PrefectHTTPStatusError as exc:
            if exc.response.status_code == 404:
                log.info(f"{item_to_delete} was already deleted")
            else:
                raise

    if trigger_type:
        log.info(
            f"Processed triggers of type {trigger_type.value}: "
            f"{len(report.created)} created, {len(report.updated)} updated, {len(report.unchanged)} unchanged, {len(report.deleted)} deleted"
        )
    else:
        log.info(
            f"Processed all triggers: "
            f"{len(report.created)} created, {len(report.updated)} updated, {len(report.unchanged)} unchanged, {len(report.deleted)} deleted"
        )

    return report



from ...generic_schema import GenericSchema

core_node = GenericSchema(
    name="Node",
    namespace="Core",
    include_in_menu=False,
    description="Base Node in Infrahub.",
    label="Node",
)

core_task_target = GenericSchema(
    name="TaskTarget",
    include_in_menu=False,
    namespace="Core",
    description="Extend a node to be associated with tasks",
    label="Task Target",
)



from typing import Any

from fast_depends import Depends, inject
from infrahub_sdk.client import InfrahubClient
from infrahub_sdk.config import Config

from infrahub import config
from infrahub.components import ComponentType
from infrahub.constants.environment import INSTALLATION_TYPE
from infrahub.core.registry import registry
from infrahub.database import InfrahubDatabase, get_db
from infrahub.services.adapters.cache import InfrahubCache
from infrahub.services.adapters.event import InfrahubEventService
from infrahub.services.adapters.http import InfrahubHTTP
from infrahub.services.adapters.http.httpx import HttpxAdapter
from infrahub.services.adapters.message_bus import InfrahubMessageBus
from infrahub.services.adapters.workflow import InfrahubWorkflow
from infrahub.services.adapters.workflow.local import WorkflowLocalExecution
from infrahub.services.adapters.workflow.worker import WorkflowWorkerExecution
from infrahub.services.component import InfrahubComponent

_singletons: dict[str, Any] = {}


def set_component_type(component_type: ComponentType) -> None:
    if "component_type" not in _singletons:
        _singletons["component_type"] = component_type


def get_component_type() -> ComponentType:
    try:
        return _singletons["component_type"]
    except KeyError as exc:
        raise ValueError("Component type is not set. It needs to be initialized before working with services.") from exc


def build_client() -> InfrahubClient:
    client_config = Config(address=config.SETTINGS.main.internal_address, retry_on_failure=True)
    client_config.set_ssl_context(context=get_http().verify_tls())
    client = InfrahubClient(config=client_config)
    # Populate client schema cache using our internal schema cache
    if registry.schema:
        for branch in registry.schema.get_branches():
            client.schema.set_cache(schema=registry.schema.get_sdk_schema_branch(name=branch), branch=branch)

    return client


@inject
def get_client(client: InfrahubClient = Depends(build_client)) -> InfrahubClient:  # noqa: B008
    return client


def build_installation_type() -> str:
    return INSTALLATION_TYPE


@inject
def get_installation_type(installation_type: str = Depends(build_installation_type)) -> str:
    return installation_type


async def build_database() -> InfrahubDatabase:
    if "database" not in _singletons:
        _singletons["database"] = InfrahubDatabase(driver=await get_db(retry=5))
    return _singletons["database"]


@inject
async def get_database(database: InfrahubDatabase = Depends(build_database)) -> InfrahubDatabase:  # noqa: B008
    return database


async def build_cache() -> InfrahubCache:
    if "cache" not in _singletons:
        _singletons["cache"] = config.OVERRIDE.cache or await InfrahubCache.new_from_driver(
            driver=config.SETTINGS.cache.driver
        )
    return _singletons["cache"]


@inject
async def get_cache(cache: InfrahubCache = Depends(build_cache)) -> InfrahubCache:  # noqa: B008
    return cache


async def build_message_bus() -> InfrahubMessageBus:
    if "message_bus" not in _singletons:
        _singletons["message_bus"] = config.OVERRIDE.message_bus or (
            await InfrahubMessageBus.new_from_driver(
                component_type=get_component_type(), driver=config.SETTINGS.broker.driver
            )
        )
    return _singletons["message_bus"]


@inject
async def get_message_bus(message_bus: InfrahubMessageBus = Depends(build_message_bus)) -> InfrahubMessageBus:  # noqa: B008
    return message_bus


async def build_event_service() -> InfrahubEventService:
    if "event_service" not in _singletons:
        _singletons["event_service"] = InfrahubEventService(message_bus=await get_message_bus())
    return _singletons["event_service"]


@inject
async def get_event_service(event_service: InfrahubEventService = Depends(build_event_service)) -> InfrahubEventService:  # noqa: B008
    return event_service


def build_workflow() -> InfrahubWorkflow:
    if "workflow" not in _singletons:
        _singletons["workflow"] = config.OVERRIDE.workflow or (
            WorkflowWorkerExecution()
            if config.SETTINGS.workflow.driver == config.WorkflowDriver.WORKER
            else WorkflowLocalExecution()
        )
    return _singletons["workflow"]


@inject
def get_workflow(workflow: InfrahubWorkflow = Depends(build_workflow)) -> InfrahubWorkflow:  # noqa: B008
    return workflow


def build_http_service() -> InfrahubHTTP:
    if "http_service" not in _singletons:
        _singletons["http_service"] = HttpxAdapter()
    return _singletons["http_service"]


@inject
def get_http(http_service: InfrahubHTTP = Depends(build_http_service)) -> InfrahubHTTP:  # noqa: B008
    return http_service


async def build_component() -> InfrahubComponent:
    if "component" not in _singletons:
        _singletons["component"] = await InfrahubComponent.new(
            cache=await get_cache(),
            component_type=get_component_type(),
            db=await get_database(),
            message_bus=await get_message_bus(),
        )
    return _singletons["component"]


@inject
async def get_component(component: InfrahubComponent = Depends(build_component)) -> InfrahubComponent:  # noqa: B008
    return component



from __future__ import annotations

from prefect import flow
from prefect.logging import get_run_logger

from infrahub import config
from infrahub.core.registry import registry
from infrahub.core.timestamp import Timestamp
from infrahub.lock import LOCK_PREFIX
from infrahub.services import InfrahubServices  # noqa: TC001  needed for prefect flow


@flow(
    name="clean-up-deadlocks",
    flow_run_name="Clean up deadlocks",
)
async def clean_up_deadlocks(service: InfrahubServices) -> None:
    """Remove stale distributed locks left behind by inactive workers"""
    keys = await service.cache.list_keys(filter_pattern=f"{LOCK_PREFIX}*")
    if not keys:
        return

    log = get_run_logger()
    values = await service.cache.get_values(keys=keys)
    workers = await service.component.list_workers(branch=registry.default_branch, schema_hash=False)
    workers_active = {worker.id for worker in workers if worker.active}

    for key, value in zip(keys, values, strict=False):
        if not key or not value:
            continue

        timestamp, worker_id = value.split("::", 1)
        if worker_id not in workers_active and Timestamp() > Timestamp(timestamp).add(
            minutes=config.SETTINGS.cache.clean_up_deadlocks_interval_mins
        ):
            await service.cache.delete(key)
            log.info(f"Deleted deadlock key={key} worker={worker_id}")



from enum import Enum


class ConstraintIdentifier(str, Enum):
    ATTRIBUTE_PARAMETERS_REGEX_UPDATE = "attribute.parameters.regex.update"
    ATTRIBUTE_PARAMETERS_MIN_LENGTH_UPDATE = "attribute.parameters.min_length.update"
    ATTRIBUTE_PARAMETERS_MAX_LENGTH_UPDATE = "attribute.parameters.max_length.update"
    ATTRIBUTE_PARAMETERS_MIN_VALUE_UPDATE = "attribute.parameters.min_value.update"
    ATTRIBUTE_PARAMETERS_MAX_VALUE_UPDATE = "attribute.parameters.max_value.update"
    ATTRIBUTE_PARAMETERS_EXCLUDED_VALUES_UPDATE = "attribute.parameters.excluded_values.update"
    ATTRIBUTE_PARAMETERS_END_RANGE_UPDATE = "attribute.parameters.end_range.update"
    ATTRIBUTE_PARAMETERS_START_RANGE_UPDATE = "attribute.parameters.start_range.update"



import ast


def get_class_def_index(module: ast.Module) -> int:
    """Get the index of the first class definition in the module.
    It's useful to insert other classes before the first class definition."""
    for idx, item in enumerate(module.body):
        if isinstance(item, ast.ClassDef):
            return idx
    return -1


def insert_fragments_inline(module: ast.Module, fragment: ast.Module) -> ast.Module:
    """Insert the Pydantic classes for the fragments inline into the module.

    If no class definitions exist in module, fragments are appended to the end.
    """
    module_class_def_index = get_class_def_index(module)

    fragment_classes: list[ast.ClassDef] = [item for item in fragment.body if isinstance(item, ast.ClassDef)]

    # Handle edge case when no class definitions exist
    if module_class_def_index == -1:
        # Append fragments to the end of the module
        module.body.extend(fragment_classes)
    else:
        # Insert fragments before the first class definition
        for idx, item in enumerate(fragment_classes):
            module.body.insert(module_class_def_index + idx, item)

    return module


def remove_fragment_import(module: ast.Module) -> ast.Module:
    """Remove the fragment import from the module."""
    for item in module.body:
        if isinstance(item, ast.ImportFrom) and item.module == "fragments":
            module.body.remove(item)
            return module
    return module
